from typing import Union

from prompt_toolkit.formatted_text import split_lines, to_formatted_text
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import (
    Container,
    Dimension,
    FormattedTextControl,
    HSplit,
    Window,
)
from prompt_toolkit.layout.containers import ConditionalContainer
from prompt_toolkit.layout.margins import ScrollbarMargin
from prompt_toolkit.layout.screen import Point


class DisplayPane:
    def __init__(
        self,
        get_formatted_text,
        filter=None,  # noqa: A002
        title="",
        height=None,
    ) -> None:
        self.cursor = Point(0, 0)
        self.get_formatted_text = get_formatted_text
        self.container: Union[HSplit, ConditionalContainer] = HSplit(
            [
                Window(
                    content=FormattedTextControl(
                        text=to_formatted_text(title, style="bold reverse"),
                        focusable=False,
                    ),
                    style="class:pane-title",
                    height=Dimension.exact(1),
                    cursorline=True,
                    always_hide_cursor=True,
                    wrap_lines=False,
                ),
                Window(
                    content=FormattedTextControl(
                        text=get_formatted_text,
                        focusable=True,
                        key_bindings=self._get_key_bindings(),
                        get_cursor_position=self._cursor_position,
                    ),
                    height=height,
                    style="class:details-box",
                    always_hide_cursor=False,
                    right_margins=[
                        ScrollbarMargin(display_arrows=False),
                    ],
                    wrap_lines=True,
                ),
            ],
            width=Dimension(preferred=80),
        )

        if filter is not None:
            self.container = ConditionalContainer(content=self.container, filter=filter)

    def _cursor_position(self):
        # don't allow the cursor to go beyond the rendered content length
        lines = list(split_lines(self.get_formatted_text()()))
        if self.cursor.y >= len(lines) - 1:
            self.cursor = Point(self.cursor.x, len(lines) - 1)
        elif self.cursor.y < 0:
            self.cursor = Point(self.cursor.x, 0)
        return self.cursor

    def _get_key_bindings(self) -> KeyBindings:
        kb = KeyBindings()

        @kb.add("up")
        def _go_up(event) -> None:
            if self.cursor.y > 0:
                self.cursor = Point(self.cursor.x, self.cursor.y - 1)

        @kb.add("down")
        def _go_down(event) -> None:
            self.cursor = Point(self.cursor.x, self.cursor.y + 1)

        return kb

    def __pt_container__(self) -> Container:
        return self.container



from __future__ import annotations

import logging
import os
from dataclasses import dataclass

from dataclasses_json import dataclass_json

from yardstick.store import config as store_config

IMAGE_LINEAGE_DIR = os.path.join("image-lineage")
SUFFIX = ".json"


@dataclass_json
@dataclass(frozen=True, eq=True, order=True)
class ImageLineageDocument:
    lineage: dict[str, list[str]]


def store_path(suffix: str = SUFFIX, store_root: str | None = None) -> str:
    if not store_root:
        store_root = store_config.get().store_root

    return os.path.join(store_root, IMAGE_LINEAGE_DIR, "image-lineage" + suffix)


def add(image: str, lineage: list[str], store_root: str | None = None):
    data_path = store_path(store_root=store_root)
    logging.debug(f"storing image lineage to {data_path!r}")

    os.makedirs(os.path.dirname(data_path), exist_ok=True)

    existing = load(store_root=store_root)
    existing[image] = lineage

    with open(data_path, "w", encoding="utf-8") as data_file:
        data_file.write(
            ImageLineageDocument(lineage=existing).to_json(indent=2),  # type: ignore[attr-defined]
        )


def get_parents(image: str, store_root: str | None = None) -> list[str]:
    return load(store_root).get(image, [])


def get(image: str, store_root: str | None = None) -> list[str]:
    result = []
    parents = get_parents(image, store_root=store_root)
    result += parents

    for parent in parents:
        for ancestor in get(parent, store_root=store_root):
            if ancestor not in result:
                result += [ancestor]
    return result


def load(store_root: str | None = None) -> dict[str, list[str]]:
    data_path = store_path(store_root=store_root)
    logging.debug(f"loading image lineage location={data_path!r}")

    if not os.path.exists(data_path):
        return {}

    with open(data_path, encoding="utf-8") as data_file:
        data_json = data_file.read()
        return ImageLineageDocument.from_json(  # type: ignore[attr-defined]
            data_json,
        ).lineage



from __future__ import annotations

import collections
import glob
import json
import logging
import os

import omitempty  # type: ignore[import]

from yardstick import artifact
from yardstick.store import config as store_config
from yardstick.store import naming

LABELS_DIR = "labels"


def label_store_root(store_root: str | None = None) -> str:
    if not store_root:
        store_root = store_config.get().store_root
    return os.path.join(store_root, LABELS_DIR)


def store_filename_by_entry(entry: artifact.LabelEntry) -> str:
    if entry.image and entry.image.exact:
        return f"{naming.image.encode(entry.image.exact)}/{entry.ID}.json"
    return f"{entry.ID}.json"


def store_path(filename: str, store_root: str | None = None) -> str:
    return os.path.join(label_store_root(store_root=store_root), filename)


def append_and_update(
    new_and_modified_entries: list[artifact.LabelEntry],
    delete_entries: list[artifact.LabelEntry] | None = None,
    store_root: str | None = None,
) -> None:
    if delete_entries:
        for label_entry in delete_entries:
            filepath = store_path(filename=store_filename_by_entry(entry=label_entry))
            try:
                logging.debug(f"deleting label {label_entry.ID} from {filepath}")
                os.remove(filepath)
            except FileNotFoundError:
                logging.debug(
                    f"skipping deleting on {label_entry.ID} from {filepath}: File not found",
                )
            except Exception as e:
                logging.error(
                    f"failed to delete label {label_entry.ID} from {filepath}: {e}",
                )
                raise e

    save(label_entries=new_and_modified_entries, store_root=store_root)


def delete(
    label_ids_to_delete: list[str],
    store_root: str | None = None,
) -> list[str]:
    """delete_entries takes a list of ids to be deleted and returns a list of deleted files.
    FileNotFound exceptions are ignored."""
    label_store_dir = label_store_root(store_root=store_root)
    deleted_ids = []
    for label_id in label_ids_to_delete:
        g = f"{label_store_dir}/**/{label_id}.json"
        for p in glob.glob(g):
            try:
                os.remove(p)
                deleted_ids.append(label_id)
            except FileNotFoundError:
                logging.debug(
                    f"skipping deleting on {label_id} from {p}: File not found",
                )
            except Exception as e:
                logging.error(f"failed to delete label {label_id} from {p}: {e}")
                raise e
    return deleted_ids


def save(label_entries: list[artifact.LabelEntry], store_root: str | None = None):
    root_path = label_store_root(store_root=store_root)
    logging.debug(f"storing all labels location={root_path}")

    # organize labels into correct destinations
    label_entries_by_destination = collections.defaultdict(set)

    for label_entry in label_entries:
        if not label_entry:
            # don't keep empty labels
            continue

        if not isinstance(label_entry, artifact.LabelEntry):
            raise RuntimeError(
                f"only LabelEntry is supported, given {type(label_entry)}",
            )

        path = store_path(filename=store_filename_by_entry(entry=label_entry))
        label_entries_by_destination[path].add(label_entry)

    for path, destination_label_entries in label_entries_by_destination.items():
        logging.debug(f"overwriting all labels location={path}")
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w", encoding="utf-8") as data_file:
            if len(destination_label_entries) == 1:
                json.dump(
                    omitempty(
                        next(iter(destination_label_entries)).to_dict(),  # type: ignore[attr-defined]
                    ),
                    data_file,
                    sort_keys=True,
                )
            else:
                logging.debug(
                    f"writing multiple labels to {path!r}: {[label_entry.ID for label_entry in destination_label_entries]}",
                )
                data_file.write(
                    artifact.LabelEntry.schema().dump(  # type: ignore[attr-defined]
                        sorted(destination_label_entries),
                        many=True,
                    ),
                )


def load_label_file(
    filename: str,
    year_max_limit: int | None = None,
    year_from_cve_only: bool = False,
    store_root: str | None = None,
) -> list[artifact.LabelEntry]:
    # why note take a file path? in this way we control that all input/output data was derived from the same store,
    # and not another store.
    path = store_path(filename=filename, store_root=store_root)

    try:
        with open(path, encoding="utf-8") as data_file:
            data_json = data_file.read()
            if not data_json.strip():
                raise FileNotFoundError

            # dataclass_json does not play nice with method inference here
            try:
                label_entries = [
                    artifact.LabelEntry.from_json(data_json),  # type: ignore[attr-defined]
                ]
            except:  # noqa: E722
                label_entries = artifact.LabelEntry.schema().load(  # type: ignore[attr-defined]
                    data_json,
                    many=True,
                )
            if year_max_limit:
                label_entries = filter_by_year(
                    label_entries,
                    year_max_limit=int(year_max_limit),
                    year_from_cve_only=year_from_cve_only,
                )
            return label_entries

    except FileNotFoundError:
        return []


def load_all(
    year_max_limit: int | None = None,
    year_from_cve_only: bool = False,
    store_root: str | None = None,
) -> list[artifact.LabelEntry]:
    root_path = label_store_root(store_root=store_root)
    logging.debug(f"loading all labels (location={root_path})")

    label_entries: list[artifact.LabelEntry] = []
    files = set(
        list(glob.glob(f"{root_path}/**/**/*.json")) + list(glob.glob(f"{root_path}/**/*.json")) + list(glob.glob(f"{root_path}/*.json")),
    )
    for file in files:
        filepath = file.removeprefix(root_path + "/")
        loaded_label_entries = load_label_file(
            filepath,
            year_max_limit=year_max_limit,
            year_from_cve_only=year_from_cve_only,
            store_root=store_root,
        )
        label_entries.extend(loaded_label_entries)

    return label_entries


def load_for_image(
    images: str | list[str],
    year_max_limit: int | None = None,
    year_from_cve_only: bool = False,
    store_root: str | None = None,
) -> list[artifact.LabelEntry]:
    root_path = label_store_root(store_root=store_root)
    if isinstance(images, str):
        images = [images]

    logging.debug(f"loading labels for image (location={root_path} image={images})")

    label_entries: list[artifact.LabelEntry] = []

    # load entries that don't have specific image
    for file in glob.glob(f"{root_path}/*.json"):
        filepath = file.removeprefix(root_path + "/")
        loaded_label_entries = load_label_file(
            filepath,
            year_max_limit=year_max_limit,
            year_from_cve_only=year_from_cve_only,
            store_root=store_root,
        )
        for entry in loaded_label_entries:
            for image in images:
                if entry.matches_image(image):
                    label_entries.append(entry)

    for image in images:
        for file in set(
            list(glob.glob(f"{root_path}/{naming.image.encode(image)}/**/*.json"))
            + list(glob.glob(f"{root_path}/{naming.image.encode(image)}/*.json")),
        ):
            filename = file.removeprefix(root_path + "/")
            loaded_label_entries = load_label_file(
                filename,
                year_max_limit=year_max_limit,
                year_from_cve_only=year_from_cve_only,
                store_root=store_root,
            )
            for entry in loaded_label_entries:
                if entry.matches_image(image):
                    label_entries.append(entry)

    return label_entries


# filter_by_year filters out CVE vuln IDs above a given year. We attempt to normalize all vuln IDs to CVEs,
# but will include any if normalization fails.
def filter_by_year(
    label_entries: list[artifact.LabelEntry],
    year_max_limit: int,
    year_from_cve_only: bool = False,
) -> list[artifact.LabelEntry]:
    label_entries_copy = []

    for label_entry in label_entries:
        year = label_entry.effective_year(by_cve=year_from_cve_only)
        if (year and year <= year_max_limit) or not year:
            label_entries_copy.append(label_entry)

    return label_entries_copy



SLASH_REPLACEMENT = "+"
SUFFIX = ".json"


class image:
    @staticmethod
    def encode(img: str) -> str:
        # allow for tags to have the slash replacement, but image names still get replaced

        if ":" in img:
            fields = img.split(":")
            image_name = ":".join(fields[:-1])
            tag = fields[-1]
            image_name = image_name.replace("/", SLASH_REPLACEMENT)
            return f"{image_name}:{tag}"

        if "@" in img:
            fields = img.split("@")
            image_name = "@".join(fields[:-1])
            digest = fields[-1]
            image_name = image_name.replace("/", SLASH_REPLACEMENT)
            return f"{image_name}@{digest}"

        return img.replace("/", SLASH_REPLACEMENT)

    @staticmethod
    def decode(name: str) -> str:
        # allow for tags to have the slash replacement, but image names still get replaced

        if ":" in name:
            fields = name.split(":")
            image_name = ":".join(fields[:-1])
            tag = fields[-1]
            image_name = image_name.replace(SLASH_REPLACEMENT, "/")
            return f"{image_name}:{tag}"

        if "@" in name:
            fields = name.split("@")
            image_name = "@".join(fields[:-1])
            digest = fields[-1]
            image_name = image_name.replace(SLASH_REPLACEMENT, "/")
            return f"{image_name}@{digest}"

        return name.replace(SLASH_REPLACEMENT, "/")



import colorsys
import json
from typing import Any, Union

from tabulate import tabulate

from yardstick import artifact, comparison


def summarize(comp: Union[comparison.ByPreservedMatch, comparison.ByMatch]):
    print(str(comp))


def preserved_matches(
    comp: comparison.ByPreservedMatch,
    details=True,
    summary=True,
    common=True,
):
    if details:
        if common:
            for common_match in comp.common:
                print("common ", common_match)
            print()

        all_rows: list[list[Any]] = []
        for result in comp.results:
            for unique_match in comp.unique[result.ID]:
                all_rows.append(
                    [
                        f"{result.config.tool_name}-{result.config.tool_version}-only",
                        unique_match.package.name,
                        unique_match.package.version,
                        unique_match.vulnerability.id,
                    ],
                )

        all_rows = sorted(all_rows)

        if len(all_rows) == 0:
            print("All results are the same (no differences found)\n")
        else:
            print(tabulate(all_rows, tablefmt="plain"))

    if summary:
        summarize(comp)


def matches(comp: comparison.ByMatch, details=True, summary=True, common=True):
    if details:
        if common:
            for match in comp.common:
                print("common ", match)

        all_rows: list[list[Any]] = []
        for result in comp.results:
            for match in comp.unique[result.ID]:
                all_rows.append(
                    [
                        f"{result.config.tool_name}-{result.config.tool_version}-only",
                        match.package.name,
                        match.package.version,
                        match.vulnerability.id,
                    ],
                )

        all_rows = sorted(all_rows)
        print(tabulate(all_rows, tablefmt="plain"))

    if summary:
        summarize(comp)


#############################################################################################################
# For label comparisons


def red(text: str) -> str:
    return f"\033[91m{text}\033[0m"


def rgb_ansi(r, g, b):
    return f"\033[38;2;{r};{g};{b}m"


def reset_ansi():
    return "\033[0m"


def get_section_rgb_tuple(index, sections):
    half_sections = int(sections / 2)
    red_hsv_tuples = list(
        reversed(
            [(0, float(x) / float(half_sections - 1), 1) for x in range(half_sections)],
        ),
    )
    green_hsv_tuples = [(0.33, float(x) / float(half_sections - 1), 1) for x in range(half_sections)]
    spectrum = red_hsv_tuples + green_hsv_tuples
    values = [colorsys.hsv_to_rgb(*x) for x in spectrum][index]
    return values[0] * 255, values[1] * 255, values[2] * 255


def get_section_index(value, min_value, max_value, sections, invert):
    value = min(max(value, min_value), max_value)
    value_ratio = float(value - min_value) / float(max_value - min_value)
    if invert:
        value_ratio = 1.0 - value_ratio
    return min(max(int(sections * value_ratio), 0), sections - 1), value_ratio


def format_value_red_green_spectrum(
    value,
    min_value=0,
    max_value=1,
    sections=10,
    invert=False,
):
    index, value_ratio = get_section_index(
        value,
        min_value,
        max_value,
        sections,
        invert,
    )
    color_rgb_tuple = get_section_rgb_tuple(index, sections)

    formatted_value = f"{rgb_ansi(*color_rgb_tuple)}{value:6.2f}{reset_ansi()}"

    if value_ratio > 0.9:
        # bold
        formatted_value = "\033[1m" + formatted_value
    if value_ratio > 0.95:
        # underline
        formatted_value = "\033[4m" + formatted_value
    return formatted_value


def format_percent(value: float) -> str:
    if value < 0:
        return ""
    return (
        "("
        + format_value_red_green_spectrum(
            value,
            min_value=0,
            max_value=50,
            invert=True,
        )
        + " %)"
    )


def show_label_comparison_summary(  # noqa: C901
    stats_by_image_tool_pair: comparison.ImageToolLabelStats,
):
    tools = stats_by_image_tool_pair.tools
    images = stats_by_image_tool_pair.images

    def summarize_by_tool(  # noqa: PLR0913
        title,
        tp,
        fp,
        fn,
        intedeterminate,
        intedeterminate_percent,
        f1,
        f1_ranges,
    ):
        header = ["", "TP", "FP", "FN", "Indeterminate", "F1 Score", "F1 Score Range"]
        all_rows = []
        for tool in sorted(tools):
            f1r = ""
            if f1_ranges.get(tool, None) and f1.get(tool, -1) > 0:
                f1r = f"{f1_ranges[tool][0]:0.2f}-{f1_ranges[tool][1]:0.2f}"

            f1s = ""
            if f1.get(tool, -1) > 0:
                f1s = f"{format_value_red_green_spectrum(f1.get(tool))}"

            row = [
                tool,
                tp.get(tool, ""),
                fp.get(tool, ""),
                fn.get(tool, ""),
                f"{intedeterminate.get(tool, '')} {format_percent(intedeterminate_percent.get(tool, -1))}",
                f1s,
                f1r,
            ]
            all_rows.append(row)

        print(f"~~~ {title!r} Summary (by Tool) ~~~")
        print(tabulate(all_rows, tablefmt="simple", headers=header))

    def summarize_across_images(title, description, source, context_source=None):
        header = ["", *sorted(tools)]
        all_rows = []
        for image in sorted(i for i in images):
            row = [image]
            for tool in sorted(tools):
                # why \0? to prevent from tabulate from stripping the whitespace
                cell = f"\0{source[image].get(tool, ''):-5}"
                if context_source:
                    cell += f" {context_source[image].get(tool, None):7}"
                row.append(cell)

            all_rows.append(row)

        print(f"\n\n\n~~~ {title} Summary ~~~")
        if description:
            print(f"\n{description}")
        print(tabulate(all_rows, tablefmt="simple", headers=header))

    for image in images:
        summarize_by_tool(
            image,
            stats_by_image_tool_pair.true_positives[image],
            stats_by_image_tool_pair.false_positives[image],
            stats_by_image_tool_pair.false_negatives[image],
            stats_by_image_tool_pair.indeterminate[image],
            stats_by_image_tool_pair.indeterminate_percent[image],
            stats_by_image_tool_pair.f1_scores[image],
            stats_by_image_tool_pair.f1_score_ranges[image],
        )

    summarize_across_images(
        "False Negative",
        """\
Ideally FNs for a vuln scanner is 0, which means that all possible known vulnerabilities that could have been
reported were (relative to the label data). So each FN is a missed vulnerability match for the tool-image pair.

Each FN entry for each tool-image pair is logged above.""",
        stats_by_image_tool_pair.false_negatives,
    )

    summarize_across_images(
        "Indeterminate Label",
        """\
Ideally there should be no indeterminate labels. The more indeterminate labels there are the wider the F1
score range is. An indeterminate label is either an explicit label of "?" or there is more that one label
for a match that is in conflict (there is a TP and a FP).

Each indeterminate match for each tool-image pair is logged above.""",
        stats_by_image_tool_pair.indeterminate,
        context_source={
            image: {tool: format_percent(value) for tool, value in values.items()}
            for image, values in stats_by_image_tool_pair.indeterminate_percent.items()
        },
    )

    # F1 Summary
    header = ["", *sorted(tools)]
    all_rows = []
    for image in sorted(i for i in images):
        row = [image]
        for tool in sorted(tools):
            f1_score = stats_by_image_tool_pair.f1_scores[image].get(tool, None)

            if f1_score and f1_score > 0:
                lower, upper = stats_by_image_tool_pair.f1_score_ranges[image].get(
                    tool,
                    (-1, -1),
                )
                f1_score_str = "error!" if lower == -1 or upper == -1 else f"{format_value_red_green_spectrum(f1_score)} ({lower:0.2f}-{upper:0.2f})"
            elif f1_score and f1_score < 0:
                lower, upper = stats_by_image_tool_pair.f1_score_ranges[image].get(
                    tool,
                    (-1, -1),
                )
                f1_score_str = "error!" if lower == -1 or upper == -1 else red(f"Impractical ({lower:0.2f}-{upper:0.2f})")
            else:
                f1_score_str = ""
            row.append(f1_score_str)

        all_rows.append(row)

    print("\n\n~~~ F1 Score Summary ~~~\n")
    print(
        """\
Ideally the F1 score for an image-tool pair should be 1. F1 score combines the TP, FP, and FN counts into a
single metric between 0 and 1. This helps summarize the matching performance but does not explain why the
matching performance is what it is.

Also, each F1 score is given a possible range based on how much label data is available. The more data that
is available, the smaller the range. Large ranges are not evaluated and the F1 score is overall deemed
Impractical.

See the TP/FP/FN counts for the tool-image sections logged above for more context.""",
    )
    print(tabulate(all_rows, tablefmt="simple", headers=header))


def label_comparison_json(
    results: list[artifact.ScanResult],
    comparisons_by_result_id: dict[str, comparison.AgainstLabels],
    stats_by_image_tool_pair: comparison.ImageToolLabelStats,
    show_fns: bool,
    show_indeterminates: bool,
):
    ret = []

    for result in results:
        comp = comparisons_by_result_id[result.ID]
        image = comp.config.image
        tool = comp.config.tool

        more = {}

        if show_fns:
            more["fns"] = [
                label.to_dict()  # type: ignore[attr-defined]
                for label in comp.false_negative_label_entries
            ]

        if show_indeterminates:
            more["indeterminate"] = []
            for match in comp.matches_with_indeterminate_labels:
                more["indeterminate"].append(
                    {
                        "match": match.to_dict(),  # type: ignore[attr-defined]
                        "label_set": sorted(
                            [label.display for label in set(comp.labels_by_match.get(match.ID, []))],
                        ),
                        "labels": [
                            label.to_dict()  # type: ignore[attr-defined]
                            for label in comp.label_entries_by_match.get(match.ID, [])
                        ],
                    },
                )

        ret.append(
            {
                "image": image,
                "tool": tool,
                "stats": {
                    "f1_score": stats_by_image_tool_pair.f1_scores[image][tool],
                    "f1_score_range": stats_by_image_tool_pair.f1_score_ranges[image][tool],
                    "fn": stats_by_image_tool_pair.false_negatives[image][tool],
                    "tp": stats_by_image_tool_pair.true_positives[image][tool],
                    "fp": stats_by_image_tool_pair.false_positives[image][tool],
                    "indeterminate": stats_by_image_tool_pair.indeterminate[image][tool],
                    "indeterminate_percent": stats_by_image_tool_pair.indeterminate_percent[image][tool],
                },
                **more,
            },
        )

    print(json.dumps(ret, indent=2))


def label_comparison(  # noqa: C901, PLR0913
    results: list[artifact.ScanResult],
    comparisons_by_result_id: dict[str, comparison.AgainstLabels],
    stats_by_image_tool_pair: comparison.ImageToolLabelStats,
    show_fns: bool = False,
    show_indeterminates: bool = False,
    show_summaries: bool = True,
):
    for result in results:
        comp = comparisons_by_result_id[result.ID]

        # show the results
        if show_summaries:
            print(str(comp))
            print()

        if show_fns:
            print(
                "False Negative Label Entries: ",
                len(comp.false_negative_label_entries),
            )
            for label in comp.false_negative_label_entries:
                print("   ", label.summarize())
            print()

        if show_indeterminates:
            print(
                "Indeterminate Matches: ",
                len(comp.matches_with_indeterminate_labels),
            )
            for match in comp.matches_with_indeterminate_labels:
                if not match.config:
                    continue
                print(
                    "   ",
                    match,
                    f"from {match.config.tool_name}@{match.config.tool_version}",
                )
                match_labels = comp.label_entries_by_match.get(match.ID, [])
                if match_labels:
                    print(
                        "    Label Set: ",
                        set(comp.labels_by_match.get(match.ID, [])),
                    )
                for label in match_labels:
                    print("      ", label.summarize(), "\n")
                if not match_labels:
                    print("      [no labels paired]\n")
            print()

        if show_summaries or show_fns or show_indeterminates:
            print("\n\n")

    if show_summaries:
        show_label_comparison_summary(stats_by_image_tool_pair)


def labels_by_ecosystem_comparison(
    results_by_image: dict[str, list[artifact.ScanResult]],
    stats: comparison.ToolLabelStatsByEcosystem,
    show_images_used: bool = True,
):
    if show_images_used:
        print("Images used:")
        for image in results_by_image:
            print(f"  {image}")

    # show table per-tool... rows are ecosystems and columns are TPs, FPs, Precision
    for tool in stats.tools:
        header = [
            "",
            "TPs",
            "FPs",
            "Precision",
        ]
        all_rows = []
        for e in stats.ecosystems:
            row = [
                e,
                stats.tps_by_tool_by_ecosystem[tool][e],
                stats.fps_by_tool_by_ecosystem[tool][e],
                format_value_red_green_spectrum(
                    stats.precision_by_tool_by_ecosystem[tool][e],
                ),
            ]
            all_rows.append(row)

        tps = sum(
            [stats.tps_by_tool_by_ecosystem[tool][e] for e in stats.ecosystems],
        )
        fps = sum(
            [stats.fps_by_tool_by_ecosystem[tool][e] for e in stats.ecosystems],
        )
        d = tps + fps
        precision = 0.0
        if d:
            precision = float(tps) / float(d)
        all_rows.append(
            [
                "[overall] --->",
                tps,
                fps,
                format_value_red_green_spectrum(precision),
            ],
        )

        print(f"\n~~~ {tool!r} Ecosystem Breakdown (across all images) ~~~")
        print(tabulate(all_rows, tablefmt="simple", headers=header))
        print()

    print(
        """\
Note: precision is inversely proportional to the false positive rate, so the lower the precision the
more false positives are negatively affecting the results.
""",
    )



import pytest
from yardstick import artifact
from yardstick.label import (
    _contains_as_value,
    find_labels_for_match,
    merge_label_entries,
)


class TestContainsAsValue:
    @pytest.fixture()
    def nested_dict(self):
        return {
            "temparature": "50",
            "treatment": {
                "second": 5,
                "last": "data",
            },
        }

    @pytest.mark.parametrize(
        ("value", "expected_path"),
        [
            ("data", True),
            (5, True),
            ("ata", False),
            ("dat", False),
            ("DATA", False),
            ("anotherdata", False),
            (10, False),
            ("temparature", False),
        ],
    )
    def test_nested_dict(self, nested_dict, value, expected_path):
        assert expected_path == _contains_as_value(nested_dict, value)

    @pytest.fixture()
    def list_of_dicts(self):
        return [{"treatment_plan": [[4, 5, 4], [4, "data", 4], [5, 5, 5]]}]

    @pytest.mark.parametrize(
        ("value", "expected_path"),
        [
            ("data", True),
            (5, True),
            (4, True),
            ("ata", False),
            ("dat", False),
            ("DATA", False),
            ("anotherdata", False),
            (10, False),
            ("treatment_plan", False),
        ],
    )
    def test_list_of_dicts(self, list_of_dicts, value, expected_path):
        assert expected_path == _contains_as_value(list_of_dicts, value)


class TestFindLabelsForMatch:
    @pytest.fixture()
    def label_entries(self):
        return [
            ##############################################################
            # CVE CVE-2020-0001
            # vat import
            artifact.LabelEntry(
                label=artifact.Label.FalsePositive,
                vulnerability_id="CVE-2020-0001",
                source="vat-import",
                user="somebody",
                package=artifact.Package(name="package", version="1.0"),
                ID="1",
            ),
            # manual entry
            artifact.LabelEntry(
                label=artifact.Label.FalsePositive,
                image=artifact.ImageSpecifier(exact="my/image:latest"),
                vulnerability_id="CVE-2020-0001",
                package=artifact.Package(name="package", version="1.0"),
                source="manual",
                user="somebody",
                ID="2",
            ),
            # manual entry, image pattern missing
            artifact.LabelEntry(
                label=artifact.Label.FalsePositive,
                vulnerability_id="CVE-2020-0001",
                package=artifact.Package(name="package", version="1.0"),
                source="manual",
                user="somebody",
                ID="3",
            ),
            #################################################################
            # CVE CVE-2020-0002
            # manual entry, with image + extra match fields
            artifact.LabelEntry(
                label=artifact.Label.FalsePositive,
                image=artifact.ImageSpecifier(exact="my/image:latest"),
                vulnerability_id="CVE-2020-0002",
                package=artifact.Package(name="package", version="1.0"),
                source="manual",
                user="somebody",
                fullentry_fields=[
                    "4-best-match-4",
                ],
                ID="4",
            ),
            # manual entry, extra match fields
            artifact.LabelEntry(
                label=artifact.Label.FalsePositive,
                vulnerability_id="CVE-2020-0002",
                package=artifact.Package(name="package", version="1.0"),
                source="manual",
                user="somebody",
                fullentry_fields=[
                    "5-best-match-5",
                    "5-another-match-5",
                ],
                ID="5",
            ),
            #################################################################
            # never match against these
            # mismatch package name
            artifact.LabelEntry(
                label=artifact.Label.FalsePositive,
                vulnerability_id="CVE-2020-0001",
                package=artifact.Package(name="never-match", version="1.0"),
                source="manual",
                user="somebody",
                fullentry_fields=[
                    "never-match",
                    "never-match",
                ],
                ID="1000",
            ),
            # different CVE
            artifact.LabelEntry(
                label=artifact.Label.FalsePositive,
                vulnerability_id="CVE-2020-11111",
                source="vat-import",
                user="somebody",
                ID="1001",
            ),
        ]

    @pytest.mark.parametrize(
        ("image", "match", "expected_label_ids"),
        [
            # case: match on all
            (
                "my/image:latest",
                artifact.Match(
                    vulnerability=artifact.Vulnerability(id="CVE-2020-0001"),
                    package=artifact.Package(name="package", version="1.0"),
                ),
                ["1", "2", "3"],
            ),
            # case: match with different image
            (
                "another-mismatched/image:latest",
                artifact.Match(
                    vulnerability=artifact.Vulnerability(id="CVE-2020-0001"),
                    package=artifact.Package(name="package", version="1.0"),
                ),
                ["1", "3"],
            ),
            # case: match with extra info
            (
                "my/image:latest",
                artifact.Match(
                    vulnerability=artifact.Vulnerability(id="CVE-2020-0002"),
                    package=artifact.Package(name="package", version="1.0"),
                    fullentry={"something": "4-best-match-4"},
                ),
                ["4"],
            ),
            # case: match with multiple extra info
            (
                "my/image:latest",
                artifact.Match(
                    vulnerability=artifact.Vulnerability(id="CVE-2020-0002"),
                    package=artifact.Package(name="package", version="1.0"),
                    fullentry={
                        "something": "5-best-match-5",
                        "another": [
                            1,
                            {
                                "weird": "5-another-match-5",
                            },
                        ],
                    },
                ),
                ["5"],
            ),
        ],
    )
    def test_find_labels_for_match(
        self,
        label_entries,
        image,
        match,
        expected_label_ids,
    ):
        ids = [m.ID for m in find_labels_for_match(image, match, label_entries)]
        assert expected_label_ids == ids


class TestMergeLabelEntries:
    @pytest.fixture()
    def existing_label_entries(self):
        return [
            artifact.LabelEntry(
                label=artifact.Label.FalsePositive,
                image="dontcare",
                vulnerability_id="CVE-2020-0001",
                source="vat-import",
                user="somebody",
                package=artifact.Package(name="package", version="1.0"),
                ID="1",
            ),
            artifact.LabelEntry(
                label=artifact.Label.FalsePositive,
                image=artifact.ImageSpecifier(exact="my/image:latest"),
                vulnerability_id="CVE-2020-0001",
                package=artifact.Package(name="package", version="1.0"),
                source="manual",
                user="somebody",
                ID="2",
            ),
            artifact.LabelEntry(
                label=artifact.Label.FalsePositive,
                vulnerability_id="CVE-2020-0001",
                image="dontcare",
                package=artifact.Package(name="package", version="1.0"),
                source="manual",
                user="somebody",
                ID="3",
            ),
        ]

    @pytest.mark.parametrize(
        ("new_label_entries", "deleted_label_ids", "expected_label_ids"),
        [
            # case: add a new label
            (
                [
                    artifact.LabelEntry(
                        label=artifact.Label.FalsePositive,
                        image="dontcare",
                        vulnerability_id="CVE-2020-0001",
                        package=artifact.Package(name="package", version="1.0"),
                        source="manual",
                        user="somebody",
                        ID="4",
                    ),
                ],
                [],
                ["1", "2", "3", "4"],
            ),
            # case: delete a new label
            (
                [],
                ["2"],
                ["1", "3"],
            ),
            # case: modify a label
            (
                [
                    artifact.LabelEntry(
                        label=artifact.Label.FalsePositive,
                        vulnerability_id="CVE-1995-DIFFERENT",
                        image="dontcare",
                        package=artifact.Package(name="WOOT", version="42.0"),
                        source="AUTOMATIC",
                        user="ANYBODY",
                        ID="3",
                    ),
                ],
                [],
                ["1", "2", "3"],
            ),
            # case: mix!
            (
                [
                    # modification of 3...
                    artifact.LabelEntry(
                        label=artifact.Label.FalsePositive,
                        image="dontcare",
                        vulnerability_id="CVE-1995-DIFFERENT",
                        package=artifact.Package(name="WOOT", version="42.0"),
                        source="AUTOMATIC",
                        user="ANYBODY",
                        ID="3",
                    ),
                    # addition of 4...
                    artifact.LabelEntry(
                        label=artifact.Label.FalsePositive,
                        vulnerability_id="CVE-2020-0001",
                        image="dontcare",
                        package=artifact.Package(name="package", version="1.0"),
                        source="manual",
                        user="somebody",
                        ID="4",
                    ),
                ],
                ["1"],
                ["2", "3", "4"],
            ),
        ],
    )
    def test_merge_label_entries(
        self,
        existing_label_entries,
        new_label_entries,
        deleted_label_ids,
        expected_label_ids,
    ):
        merged_entries = merge_label_entries(
            existing_label_entries,
            new_label_entries,
            deleted_label_ids,
        )

        # technically the second assertion includes this, but this is easier to see when debugging
        ids = [m.ID for m in merged_entries]
        assert expected_label_ids == ids

        # check that we persisted mutated data for all elements
        for entry in new_label_entries:
            assert entry in merged_entries
